<?php
/*
Plugin Name: Double Knot Security
Description: Stop brute force login attempts by user name.
Version: 1.0.0
Author: Kicks Digital Marketing LLC
Author URI: https://kicksdigitalmarketing.com
License: GPL2+
*/

/**
* Save default settings for plugin so that it starts working right away.
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param N/A
* @return void
=====================================================================*/
function kdm_dks_activate() {
	kdm_dks_get_settings();
}
register_activation_hook( __FILE__, 'kdm_dks_activate' );

/**
* Clean Up.
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param N/A
* @return void
=====================================================================*/
function kdm_dks_deactivate() {
	delete_option( 'kdmdks_guilty_details' );
	delete_option( 'kdmdks_guilty' );
	delete_option( 'kdmdks_logs' );
	delete_option( 'kdmdks_settings' );
	delete_option( 'kdmdks_innocent' );
}
register_deactivation_hook( __FILE__, 'kdm_dks_deactivate' );

/**
* Check for a banned IP / Stop Author Enumeration.  Set HEADER to 403 if true.
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param N/A
* @return void
=====================================================================*/
function kdm_dks_startup() {
	
	$guilty = get_option( 'kdmdks_guilty', '' );
	
	if ( $guilty_details = get_option( 'kdmdks_guilty_details', array() ) ) {
		foreach ( $guilty_details as $ip => $time ) {
			if ( $time < current_time( 'timestamp' ) ) {
				unset( $guilty_details[ $ip ] );
				$guilty = str_replace( $ip, '', $guilty );
			}
		}

		update_option( 'kdmdks_guilty_details', $guilty_details );
		update_option( 'kdmdks_guilty', $guilty );
	}
	
	if ( $guilty ) {
		if ( kdm_dks_get_ip() && in_array( kdm_dks_get_ip(), explode( "\n", $guilty ) ) ) {
			header( 'HTTP/1.0 403 Forbidden' );
			exit();	
		}
	}

	if ( isset( $_GET['author'] ) ) {
		$settings = kdm_dks_get_settings();
		if ( $settings['stop_enumeration'] && get_option( 'permalink_structure', '' ) ) {
			wp_safe_redirect( '/' );
			exit();
		}
	}
}
add_action( 'wp_loaded', 'kdm_dks_startup' );

/**
* Check if tried username meets bad username criteria.
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param $user object WP_USER or WP_ERROR object
* @param $username Submitted username
* @param $password Submitted password
* @return object WP_USER or WP_ERROR object
=====================================================================*/
function kdm_dks_check_username( $user, $username, $password ) {
	global $wpdb;
	
	$settings = kdm_dks_get_settings();
	$ip       = kdm_dks_get_ip();
	
	if ( empty( $_POST ) || ! $username || in_array( $ip, $settings['wl_ips'] ) ) {
		return $user;
	}
	
	$bl = array_map( 'strtolower', $settings['bl_usernames'] );
	if ( 'black_list' == $settings['username_check'] && in_array( strtolower( $username ), $bl ) ) {
		kdm_dks_found_guilty( $ip );
		
	} elseif ( 'current_users' == $settings['username_check'] ) {
		$usernames = $wpdb->get_col( "SELECT user_login FROM $wpdb->users" );
		if ( ! in_array( $username, $usernames ) ) {
			kdm_dks_found_guilty( $ip );
		}
	}

	if ( ! empty( $_POST ) && $username && is_wp_error( $user ) ) {
		$logs = get_option( 'kdmdks_logs', array() );
		array_unshift( $logs, array(
			'date'     => current_time( 'mysql' ),
			'username' => $username,
			'ip'       => $ip,
		) );
		update_option( 'kdmdks_logs', array_slice( $logs, 0, 30 ) );
	}
	
	return $user;
}
add_filter( 'authenticate', 'kdm_dks_check_username', 99, 3 );


/**
* Add IP to guilty list and then return 403 HEADER.
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param $ip string Offending IP
* @return void
=====================================================================*/
function kdm_dks_found_guilty( $ip ) {
	
	$settings = get_option( 'kdmdks_settings', array() );

	$guilty = get_option( 'kdmdks_guilty', '' );
	$guilty_details = get_option( 'kdmdks_guilty_details', array() );
	$guilty = '' == $guilty ? $ip : $guilty . "\n" . $ip;
	update_option( 'kdmdks_guilty', $guilty );
	
	$guilty_details[ $ip ] = current_time( 'timestamp' ) + ( intval( $settings['jail_time'] ) * 3600 );
	update_option( 'kdmdks_guilty_details', $guilty_details );
	
	header( 'HTTP/1.0 403 Forbidden' );
	exit();
}

/**
* Get user's IP
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param N/A
* @return string The IP of the user
=====================================================================*/
function kdm_dks_get_ip() {
  
	if ( isset( $_SERVER['REMOTE_ADDR'] ) )    { 
    	return $_SERVER['REMOTE_ADDR']; 
	}
	if ( isset( $_SERVER['HTTP_X_FORWARDED_FOR'] ) )    { 
		return $_SERVER['HTTP_X_FORWARDED_FOR']; 
	}
	if ( isset( $_SERVER['HTTP_CLIENT_IP'] ) )    { 
		return $_SERVER['HTTP_CLIENT_IP']; 
	} 
	
	return '';
}

/**
* Return saved or default settings for plugin.
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param N/A
* @return array Plugin Settings
=====================================================================*/
function kdm_dks_get_settings() {
	
	$settings = get_option( 'kdmdks_settings', array() );
	
	if ( ! empty( $settings ) ) {
		return $settings;	
	}
	
	$uploads    = wp_upload_dir();
	$users      = get_users();
	$default_bl = array( 'admin' );
	
	foreach ( $users as $user ) {
		if ( $i = array_search( $user->user_login, $default_bl ) ) {
			unset( $default_bl[ $i ] );	
		}
		if ( $user->user_nicename != $user->user_login ) {
			$default_bl[] = $user->user_nicename;	
		}
	}
	
	//DEFAULT PLUGIN SETTINGS
	$settings = array(
		'username_check'   => 'black_list',
		'bl_usernames'     => $default_bl,
		'wl_ips'           => array( kdm_dks_get_ip() ),
		'jail_time'        => 1,
		'stop_enumeration' => 0,
		'unblock_ip'       => ''
	);
	
	update_option( 'kdmdks_settings', $settings );
	
	return $settings;
}

/**
* Converts array of items to string with new line seperation.
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param $array array Array of items
* @return string Items seperated by a newline
=====================================================================*/
function kdm_dks_array_to_newline( $array = array() ) {
	
	if ( ! is_array( $array ) ) {
		return $array;	
	}
	
	return implode( "\n", $array );
}

/**
* Checks and saves plugins settings.
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param N/A
* @return void
=====================================================================*/
function kdm_dks_manager_updates() {
	
	//VERIFY NONCE
	if ( ! wp_verify_nonce( $_REQUEST['kdmdks-nonce'], 'kdmdks-nonce' ) ) {
		return false;	
	}
	
	if ( isset( $_GET['bl-username'] ) && ! get_user_by( 'login', $_GET['bl-username'] ) ) {
		$settings = kdm_dks_get_settings();
		$settings['bl_usernames'][] = sanitize_user( $username, trim( $_GET['bl-username'] ), true );
		$settings['bl_usernames']   = array_unique( $settings['bl_usernames'] );
		update_option( 'kdmdks_settings', $settings );
		add_action( 'admin_notices', 'kdm_dks_username_added' );
	}
	
	if ( isset( $_GET['unblock-ip'] ) && filter_var( $_GET['unblock-ip'], FILTER_VALIDATE_IP ) ) {
		if ( $wl_ips = kdm_dks_unblock_ip( $_GET['unblock-ip'] ) ) {	
			$settings = kdm_dks_get_settings();
			$settings['wl_ips'] = $wl_ips;
			update_option( 'kdmdks_settings', $settings );
			add_action( 'admin_notices', 'kdm_dks_ip_unblocked' );
		}
	}
	
	if ( isset( $_POST['kdmdks'] ) && is_array( $_POST['kdmdks'] ) ) {
		
		$settings     = kdm_dks_get_settings();
		$new_settings = $_POST['kdmdks'];
		
		//CREATE ARRAY AND SANITIZE EACH LINE
		if ( ! $new_settings['bl_usernames'] ) {
			$new_settings['bl_usernames'] = array();	
		} else {
			$new_settings['bl_usernames'] = array_unique( array_map( 'trim', explode( "\n", $new_settings['bl_usernames'] ) ) );
			$new_settings['bl_usernames'] = array_map( 'sanitize_user', $new_settings['bl_usernames'] );
		}
		
		//CREATE ARRAY AND SANITIZE EACH LINE
		if ( ! $new_settings['wl_ips'] ) {
			$new_settings['wl_ips'] = array();	
		} else {
			$new_settings['wl_ips'] = array_unique( array_map( 'trim', explode( "\n", $new_settings['wl_ips'] ) ) );
			$new_settings['wl_ips'] = array_map( 'sanitize_text_field', $new_settings['wl_ips'] );
		}
		
		//SANTIZE SETTINGS
		$new_settings['username_check']   = sanitize_text_field( $new_settings['username_check'] );
		$new_settings['jail_time']        = intval( $new_settings['jail_time'] );
		$new_settings['stop_enumeration'] = ! isset( $new_settings['stop_enumeration'] ) ? 0 : intval( $new_settings['stop_enumeration'] );
		$new_settings['unblock_ip']       = sanitize_text_field( $new_settings['unblock_ip'] );

		
		//DON'T ALLOW ACTUAL USERNAMES TO BE BLACK LISTED.
		foreach ( $new_settings['bl_usernames'] as $i => $username ) {
			if ( get_user_by( 'login', $_GET['bl-username'] ) ) {
				unset( $new_settings['bl_usernames'][ $i ] );	
			}
		}
		
		//STRIP NON-IP LINES
		foreach ( $new_settings['wl_ips'] as $i => $ip ) {
			if ( ! filter_var( $ip, FILTER_VALIDATE_IP ) ) {
				unset( $new_settings['wl_ips'][ $i ] );
			}
		}
		
		//CHECK FOR IP 
		if ( filter_var( $new_settings['unblock_ip'], FILTER_VALIDATE_IP ) ) {
			if ( $wl_ips = kdm_dks_unblock_ip( $new_settings['unblock_ip'] ) ) {
				$new_settings['wl_ips'] = $wl_ips;	
			}
		}
		update_option( 'kdmdks_settings', $new_settings );
		add_action( 'admin_notices', 'kdm_dks_settings_saved' );
	}
}
//ONLY NEED TO ADD ACTION IF IT'S OUR PAGE
if ( isset( $_GET[ 'page' ] ) && $_GET[ 'page' ] == 'kdmdks' ) {
	add_action( 'admin_init', 'kdm_dks_manager_updates' );
}

/**
* Admin notice for adding username to black list.
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param N/A
* @return array Array of White list IPs
=====================================================================*/
function kdm_dks_unblock_ip( $ip = '', $wl_ips = array() ) {
	
	if ( ! filter_var( $ip, FILTER_VALIDATE_IP ) ) {
		return $wl_ips;	
	}
	
	$settings = kdm_dks_get_settings();
	
	//REMOVE IP FROM GUILTY LIST
	$guilty = get_option( 'kdmdks_guilty', '' );
	$guilty = str_replace( $ip, '', $guilty );
	update_option( 'kdmdks_guilty', $guilty );
	
	$guilty_details = get_option( 'kdmdks_guilty_details', array() );
	if ( isset( $guilty_details[ $ip ] ) ) {
		unset( $guilty_details[ $ip ] );
		update_option( 'kdmdks_guilty_details', $guilty_details );
	}

	//ADD TO INNOCENT LIST IF NOT ALREADY ON - FOR SHELL SCRIPTS
	$innocent = get_option( 'kdmdks_innocent', '' );
	if ( strpos( $innocent, $ip ) === false ) {
		$innocent = '' == $innocent ? $ip : $innocent . "\n" . $ip;
		update_option( 'kdmdks_innocent', $innocent );
	}

	//STOP FUTURE MISTAKEN BANS
	$wl_ips[] = $ip;
	$wl_ips   = array_unique( $wl_ips );
	
	return $wl_ips;
}

/**
* Admin notice for adding username to black list.
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param N/A
* @return void
=====================================================================*/
function kdm_dks_username_added() {
	?>
    <div class="notice notice-success is-dismissible">
        <p>Username added to black list.</p>
    </div>
    <?php
}

/**
* Admin notice for unblocking an IP.
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param N/A
* @return void
=====================================================================*/
function kdm_dks_ip_unblocked() {
	?>
    <div class="notice notice-success is-dismissible">
        <p>IP Unblocked.</p>
    </div>
    <?php
}

/**
* Admin notice for saving settings
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param N/A
* @return void
=====================================================================*/
function kdm_dks_settings_saved() {
	?>
    <div class="notice notice-success is-dismissible">
        <p>Settings saved successfully.</p>
    </div>
    <?php
}

/**
* Admin Menu Action
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param N/A
* @return void
=====================================================================*/
function kdm_dks_add_menu() {
	add_options_page( 'Double Knot Security', 'Double Knot Security', 'manage_options', 'kdmdks', 'kdm_dks_manager' );
}
add_action( 'admin_menu', 'kdm_dks_add_menu' );

/**
* Enqueue Admin Styles
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param $hook string Page Value
* @return void
=====================================================================*/
function kdm_dks_admin_style( $hook ) {
	if ( $hook != 'settings_page_kdmdks' ) {
		return;
	}
	wp_enqueue_style( 'kdmdks_admin_css', plugins_url( 'kdmdks-style.css', __FILE__ ) );
}
add_action( 'admin_enqueue_scripts', 'kdm_dks_admin_style' );

/**
* Build the settings page.
*
* @author Kicks Digital Marketing 
* @since 1.0.0
*
* @param N/A
* @return void
=====================================================================*/
function kdm_dks_manager() {
	global $wpdb;
	
	$tabs = array(
		'settings' => array( 'title' => 'Settings', 'icon' => 'admin-settings' ),
		'guilty'   => array( 'title' => 'Lockouts', 'icon' => 'lock' ),
		'logs'     => array( 'title' => 'Logs', 'icon' => 'list-view' ),
		'iptables' => array( 'title' => 'Shell Scripts', 'icon' => 'editor-code' ),
	);
	
	$cur_tab  = isset( $_GET['tab'] ) && isset( $tabs[ $_GET['tab'] ] ) ? $_GET['tab'] : 'settings';
	$settings = kdm_dks_get_settings();
?>
	<a name="top"></a>
	<div class="wrap">
		<h2><?php echo __( 'Double Knot Security Settings', 'double-knot-security' ); ?></h2>
		<ul id="kdmdks-nav">
			<?php foreach ( $tabs as $id => $data ) : ?>
				<li class="<?php echo esc_attr( $cur_tab == $id ? 'active' : '' ); ?>">
                	<a href="?page=kdmdks&tab=<?php echo esc_attr( $id ); ?>">
                    	<span class="dashicons dashicons-<?php echo esc_attr( $data['icon'] ); ?>"></span> <?php echo esc_attr( $data['title'] ); ?>
                    </a>
                </li>
			<?php endforeach; ?>
		</ul>
        <div id="kdmdks-panes">
		<?php switch( $cur_tab ) : 
				case 'settings' : ?>
				<form action="" method="post">
					<?php wp_nonce_field( 'kdmdks-nonce', 'kdmdks-nonce' ); ?>
					<h3>Username Settings</h3>
					<div class="kdmdks-info">
						Checking against existing user names is the most secure way to protect your blog.  If you have many users or users that may forget his/her user name this option will block them as well.  By default "admin" and user nicenames are added to your black list to block attacks. 
					</div>
					<div class="kdmdks-field">
						<label>User Name Check</label>
						<input type="radio" name="kdmdks[username_check]" value="current_users" <?php checked( 'current_users', $settings['username_check'] ); ?> /> Check Against Current Users<br>
						<input type="radio" name="kdmdks[username_check]" value="black_list" <?php checked( 'black_list', $settings['username_check'] ); ?> /> Check Against Black List<br>
					</div>
					<div class="kdmdks-field">
						<label>Blacklist Usernames (one per line)</label>
						<textarea name="kdmdks[bl_usernames]"><?php echo esc_attr( kdm_dks_array_to_newline( $settings['bl_usernames'] ) ); ?></textarea>
					</div>
					
					<h3>IP Settings</h3>
					<div class="kdmdks-info">
						IPs like the IP of your office should go here.  This will ensure that you and your co-workers don't get locked out.
					</div>
					<div class="kdmdks-field">
						<label>White List IPs (one per line) - Your IP: <?php echo kdm_dks_get_ip(); ?></label>
						<textarea name="kdmdks[wl_ips]"><?php echo esc_attr( kdm_dks_array_to_newline( $settings['wl_ips'] ) ); ?></textarea>
					</div>
					
					<h3>General Settings</h3>
                    <div class="kdmdks-field">
						<label>Jail Time</label>
                        <div class="kdmdks-info">
                        	Set how long an IP should be banned for.  Default is 1 hour.
                        </div>
						<input type="text" name="kdmdks[jail_time]" value="<?php echo esc_attr( $settings['jail_time'] ); ?>" />
					</div>
					<div class="kdmdks-field">
						<label>Stop Author Enumeration</label>
                        <div class="kdmdks-info">
                        	Author Enumeration is when an attacker attempts to discover user names by checking if user ids exist.
                        </div>
						<input type="checkbox" name="kdmdks[stop_enumeration]" value="1" <?php checked( 1, $settings['stop_enumeration'] ); ?> /> Stop Enumeration.
					</div>
					<div class="kdmdks-field">
						<label>Unblock IP</label>
                        <div class="kdmdks-info">
                        	If an IP is blocked by mistake add it below and it will be unblocked and added to the white list.
                        </div>
						<input type="text"name="kdmdks[unblock_ip]" value="" />
					</div>
					<div class="kdmdks-field">
						<input type="submit" class="button button-primary" value="save" />
					</div>
				</form>
			<?php 
				break;
	
				case 'guilty' : ?>
				<ul class="kdmdks-lockouts">
                	<li>
                    	<div class="ip">IP</div>
                        <div class="time">LOCKED OUT UNTIL</div>
                    </li>
                    <?php if ( $ips = get_option( 'kdmdks_guilty_details', array() ) ) : ?>
						<?php foreach ( $ips as $ip => $time ) : ?>
						<li>
							<div class="ip"><?php echo esc_attr( $ip ); ?> <a href="?page=kdmdks&tab=guilty&unblock-ip=<?php echo esc_attr( $ip ); ?>&kdmdks-nonce=<?php echo wp_create_nonce( 'kdmdks-nonce' ); ?>" class="dashicons dashicons-unlock" title="Unlock"></a></div>
							<div class="time"><?php echo esc_attr( date( 'M jS g:ia', $time ) ); ?></div>
						</li>
						<?php endforeach; ?>
                    <?php else : ?>
                   		<li>No Locked Out IPs</li>
                    <?php endif; ?>
                </ul>
			<?php 
				break;
	
				case 'logs' : ?>
				<ul class="kdmdks-logs">
                	<li>
                    	<div class="ip">IP</div>
                        <div class="username">USERNAME</div>
                        <div class="time">ENTRY TIME</div>
                    </li>
                    <?php if ( $logs = get_option( 'kdmdks_logs', array() ) ) : ?>
						<?php foreach ( $logs as $entry ) : ?>
						<li>
							<div class="ip"><?php echo esc_attr( $entry['ip'] ); ?></div>
							<div class="username"><?php echo esc_attr( $entry['username'] ); ?> <a href="?page=kdmdks&tab=logs&bl-username=<?php echo esc_attr( $entry['username'] ); ?>&kdmdks-nonce=<?php echo wp_create_nonce( 'kdmdks-nonce' ); ?>" class="dashicons dashicons-flag" title="Add to Blacklist"></a></div>
							<div class="time"><?php echo esc_attr( $entry['date'] ); ?></div>
						</li>
						<?php endforeach; ?>
                    <?php else : ?>
                   		<li>Log is Empty</li>
                    <?php endif; ?>
                </ul>
			<?php break; 
		
				case 'iptables' : ?>
                <br><br>The plugin blocks guilty IPs by default after Wordpress loads from the database.  If you want to block IPs at the Apache level before PHP even loads use the scripts below.
                <h3>Cron Line</h3>
                <div class="kdmdks-info">
					Add the line below to your "crontab" to run your schell script every minute to check for attacking IPs. Replace the "{{PATH TO SHELL SCRIPT}}"  with the actual path to the shell  script you create from below.
				</div>
                <div class="code">
                * * * * * sh {{PATH TO SHELL SCRIPT}} >/dev/null 2>&1
                </div>
                
                <h3>Shell script to add IPs to iptables from the database</h3>
                <div class="kdmdks-info">
						The code below will read from your databse guilty IPs field then add them to your iptables.  It then emptys the field. Replace the "{{USERNAME}}", etc. below with the actual values for those field.
				</div>
                <div class="code">
               		 	GUILTY=$(echo "SELECT option_value FROM <?php echo esc_attr( $wpdb->options ); ?> WHERE option_name = 'kdmdks_guilty';" | mysql -D {{DOMAIN}} -h {{HOST}}  -u{{USERNAME}} -p{{PASSWORD}} -s -N )<br>
                        INNOCENT=$(echo "SELECT option_value FROM <?php echo esc_attr( $wpdb->options ); ?> WHERE option_name = 'kdmdks_innocent';" | mysql -D {{DOMAIN}} -h {{HOST}}  -u{{USERNAME}} -p{{PASSWORD}} -s -N )<br><br>

						for ip in $(echo $GUILTY | sed "s/,/ /g")<br>
						do<br>
                        	<?php if ( $settings['wl_ips'] ) : ?>
                        &nbsp; if ["$ip" != "<?php echo esc_attr( implode( '"] && ["$ip" != "', $settings['wl_ips'] ) ); ?>"] then<br>
                            <?php endif; ?>
                        &nbsp; &nbsp; sudo iptables -A INPUT -s $ip -j DROP<br>
                            <?php if ( $settings['wl_ips'] ) : ?>&nbsp; fi<br><?php endif; ?>
						done<br><br>
                        
                        for ip in $(echo $INNOCENT | sed "s/,/ /g")<br>
						do<br>
                        &nbsp; sudo iptables -D INPUT -s $ip -j DROP<br>
						done<br><br>
                        
						mysql -D {{DOMAIN}} -h {{HOST}}  -u{{USERNAME}} -p{{PASSWORD}} -e "UPDATE <?php echo esc_attr( $wpdb->options ); ?> SET option_value = '' WHERE option_name = 'kdmdks_guilty' OR option_name = 'kdmdks_innocent'"
				</div>
			<?php break; ?>
		<?php endswitch; ?>
        </div>
	</div>
<?php
}